from mesa import Agent, Model
import mesa
from mesa.datacollection import DataCollector
from mesa.time import RandomActivation
from mesa.space import MultiGrid, SingleGrid
import numpy as np
from matplotlib import pyplot as plt
import seaborn as sns
import math
import random
from tqdm import tqdm
import cv2
import stats





#******************************************************************************
#******************************************************************************
# COMPUTING ECONOMICS VARIABLES 

#==============================================================================
# Compute wage in manufacturing jobs
#
#==============================================================================
def compute_Wm():
    if (True):
        return alpha * A_m * math.pow(L_m, alpha-1)                               
    Xi2 = alpha * A_m * math.pow(eta/(1-eta), alpha * eta) * math.pow((1-eta/mu)/Zm, alpha - 1)
    return Xi2 * math.pow(L_m, alpha-1)

#==============================================================================
# Compute wage in agricultural jobs
#
#==============================================================================
def compute_Wa():
    if (True):
        Y_a = A_a * math.pow(L_a, phi)
        # print(Y_a)
        Y_m = A_m * math.pow(L_m, alpha)
        P = rho * math.pow(Y_m/Y_a, sigma)
        return phi * A_a * P * math.pow(L_a, phi-1)
    Xi3 = A_a *  math.pow(Za, 1 - phi)
    Y_a = Xi3 * math.pow(L_a, phi)
    # print(Y_a)
    Xi1 = A_m * math.pow(Zm, 1 - alpha) * math.pow(math.pow(eta/(1-eta), eta)*(1-eta/mu), alpha)  
    Y_m = Xi1 * math.pow(L_m, alpha)
    P = rho * math.pow(Y_m/Y_a, sigma)
    Xi4 = phi * A_a / math.pow(Za, phi - 1)
    return Xi4 * P * math.pow(L_a, phi-1)

#==============================================================================
# Compute optimal urban population from manufactural wage 
#
#==============================================================================
def compute_Lm_opt():
    return math.pow(alpha*A_m/W_m, 1/(1-alpha))






#******************************************************************************
#******************************************************************************
# INITIALIZATION

#==============================================================================
# Initialization of the population matrices
# origin list = random proportion of 2 for urban people
# migration list = only zeros
#
# 0: stay in the countryside 1: migrate to the city without wage 2: migrate to the city with wage
#==============================================================================
def init_list(width, height):
    global constant
    origin_matrix = np.zeros((height, width))                  # create the origin list
    migra_list = np.zeros((height, width))              # create the migration list 

    
    x_matrix = np.zeros((height, width))
    dotx_matrix = np.zeros((height, width))
    x_matrix = 2*np.random.rand(height, width)-constant
    
    weights_matrix_matrix = [[np.random.rand(height, width) * social_link for _ in range(width)] for _ in range(height)]
    for rows in weights_matrix_matrix:
        for inner_matrix in rows:
                for i in range(height):
                    for j in range(width):
                        if inner_matrix[i, j] < sparse_factor:
                            inner_matrix[i, j] = 0

    
    total = width*height                                # total population
    to_urban = int(total*L_mu)                          # total urban population
    index = random.sample(range(0, total), to_urban)    # create a list of random numbers representing the urban people
    for i in range(to_urban):                           # implementing urban people in the origin matrix 
        row = int(index[i]/width)
        col = index[i]%width
        origin_matrix[row][col] = 2
    return origin_matrix, migra_list, x_matrix, dotx_matrix, weights_matrix_matrix    # return matrices






#******************************************************************************
#******************************************************************************
# UTILITIES

#==============================================================================
# Save an image showing the entire population for step == title
# white pixel = 0 = rural
# light blue pixel = 1 = urban unemployed
# dark blue pixel = 2 = urban employed
#==============================================================================
def plot_list(data, title):
    # tips = sns.load_dataset("tips")
    # sns.jointplot(x = "total_bill", y = "tip", data = tips, kind ="hex", color="lightcoral")
    # with sns.axes_style("dark"):
    #     sns.jointplot(x, y, kind="hex")
    plt.figure(dpi = 500)
    plt.title("simulation steps: "+title)
    # plt.imshow(data, cmap='Blues')
    plt.imshow(data)
    plt.colorbar()
#    plt.savefig("./image/"+title+".png")
    plt.savefig("C:/Users/marti/Documents/ECO_THESIS/image/"+title+".png")
    plt.close()
    
#==============================================================================
# Save an image showing the evolution of "title"
# 
#==============================================================================    
def plot_line(data, title):
    y = data
    x = np.arange(0,len(data),1)
    plt.figure(dpi = 500)
    plt.plot(x,y)
    plt.xlabel("simulation steps")
    plt.ylabel(title)
    plt.title(title)
    plt.grid()
    plt.savefig(title+'.png')



#==============================================================================
# Save an image showing the evolution of "title"
# 
#==============================================================================    
def plot_line_dot(data, title, dot):
    y = data
    x = np.arange(0,len(data),1)
    plt.figure(figsize=(8,4), dpi = 500)
    plt.plot(x,y)
    plt.plot(dot, y[dot], marker='o', markersize=10)
    plt.xlabel("simulation steps")
    plt.ylabel(title)
    plt.title("simulation steps: "+str(dot))
    plt.grid()
    plt.savefig("C:/Users/marti/Documents/ECO_THESIS/image/"+title+str(dot)+".png")
    plt.close()


#==============================================================================
# Save a video showing the evolution of the entire population by step
# 
#==============================================================================  
def gen_video(length,name):
    video=cv2.VideoWriter(name+'.avi',cv2.VideoWriter_fourcc(*'MJPG'),10,(1280,720))
    if name == 'result':
        img_name = ""
        video.write(cv2.resize(cv2.imread('C:/Users/marti/Documents/ECO_THESIS/image/origin.png'),(1280,720)))
    if name == 'x':
        img_name = "x"
    if name == 'dotx':
        img_name = "dotx"
    if name == "urban_share":
        img_name = "urban_share"
    if name == "unemployment_rate":
        img_name = "unemployment_rate"
    for i in tqdm(range(0,length)):
#        img=cv2.imread('./image/'+name[i]+'.png')  #Read image
        img=cv2.imread('C:/Users/marti/Documents/ECO_THESIS/image/'+img_name+str(i)+'.png')
        # Redimensionner l'image
        img=cv2.resize(img,(1280,720))
        video.write(img)
    if name == "urban_share" or name == "unemployment_rate":
        img=cv2.imread('C:/Users/marti/Documents/ECO_THESIS/image/'+img_name+str(length)+'.png')
        # Redimensionner l'image
        img=cv2.resize(img,(1280,720))
        video.write(img)

#==============================================================================
# Save an image showing the initial population (origin matrix)
# 
#==============================================================================  
def plot_grid(data):
    plt.figure(dpi = 500)
    plt.title("purple = rural, green = unemployed, yellow = urban")
    # plt.imshow(data, cmap='Blues')
    plt.imshow(data)






#******************************************************************************
#******************************************************************************
# MESA

#==============================================================================
# MESA MODEL
# 
#==============================================================================
"""A model with some number of agents."""
class MigrationModel(Model):
    
   #==============================================================================
   # MESA MODEL initialization
   # 
   #============================================================================== 
    def __init__(self, L, width, height):
        super().__init__()
        self.num_agents = L                             # Size of population
#        self.grid = MultiGrid(width, height, False)    # Create a grid (MultiGrid = multiple agents/cell)
        self.grid = SingleGrid(width, height, False)    # Create a grid (SingleGrid = single agent/cell)
        self.schedule = RandomActivation(self)          # Scheduler will call AGENT in a random order
        self.running = True
        self._steps = 0
        self._time = 0
    
        for i in range(self.num_agents):                # Fullfill the grid by MESA AGENT
            a = Harris_Todaro_Agent(i, self)            # Initialize MESA AGENT
            self.schedule.add(a)                        # Add AGENT to schedule
            x = int(i/self.grid.width)                   
            y = i%self.grid.width                        
            self.grid.place_agent(a, (x, y))            # Add AGENT on the grid

    #==============================================================================
    # MESA MODEL uptade
    # 
    #==============================================================================
    def update_data(self):
        global migra_list
        global origin_list
        global x_matrix
        global dotx_matrix
        global L_a
        global L_m
        global L_mu
        global Lm_opt
        global L_mu_share
        global unemploy_rate
        global unemployed_tot
        global W_a_ev
        global W_m_ev
        global find_job
        global random_unemployment
        #==============================================================================
        # Compute urban share  
        # Append the list : L_mu
        L_mu = np.sum(migra_list)/(2*self.grid.width*self.grid.height)          # L_mu = urban population / total population
        L_mu_share.append(L_mu)                                                 # Append the list : L_mu_share       
        unemployed = 0
        #==============================================================================
        # If there is less urban population than manufacturing jobs (NO UNEMPLOYMENT)
        if L_mu <= Lm_opt:                                                      # If urban population is lower than optimal urban population
            origin_list = migra_list                                            # Set population list to migration list (step migration)
            L_a = 1-L_mu                                                        # Update rural population
            L_m = L_mu                                                          # All urban population is employed in manufactures
                          
        #==============================================================================
        # If there is more urban population than manufacturing jobs (UNEMPLOYMENT)
        else:
            L_a = 1-L_mu                                                        # Update rural population
            L_m = Lm_opt                                                        # Manufacturing workers < urban population

            
            total_migr = np.sum(migra_list)/2                                           # Count all people migrating or staying to urban sector
            unemployed = round(total_migr - self.grid.width*self.grid.height*Lm_opt)      # Count all people unemployed (total urban pop - total jobs)
            
            if random_unemployment: 
             
                index = random.sample(range(0, int(total_migr)), unemployed)                # Finding random urban population to be unemployed
                count = 0
                for i in range(self.grid.height):                                   # Setting unemployed urban population
                    for j in range(self.grid.width):
                        if migra_list[i][j] != 0:
                            if count in index:
                                migra_list[i][j] = 1
                            count += 1
              
            else:
                count = 0
                rural_now = []
                unemployed_still = []
                for i in range(height):                                  
                    for j in range(width):
                        if migra_list[i][j] == 0:
                            rural_now.append(count)
                        if origin_list[i][j] == 1 and migra_list[i][j] != 0:
                            if random.random()> find_job:
                                unemployed_still.append(count)
                        count += 1
                total_new_unemployed = unemployed - len(unemployed_still)
                potential_new_unemployed = set(range(width*height))
                potential_new_unemployed = potential_new_unemployed - set(rural_now) - set(unemployed_still)         
    
                if total_new_unemployed>=0:
                    index = sorted(random.sample(list(potential_new_unemployed), total_new_unemployed))                # Finding random urban population to be unemployed
                    count = 0
                    for i in range(height):                                   # Setting unemployed urban population
                        for j in range(width):
                            if migra_list[i][j] != 0:
                                if count in index or count in unemployed_still:
                                    migra_list[i][j] = 1
                            count += 1
                                                              # Set population list to migration list (step migration)     
                else :
                    index = sorted(random.sample(list(unemployed_still), unemployed))
                    count = 0
                    for i in range(height):                                   # Setting unemployed urban population
                        for j in range(width):
                            if migra_list[i][j] != 0:
                                if count in index:
                                    migra_list[i][j] = 1
                            count += 1
                                   
            origin_list = migra_list                                            # Set population list to migration list (step migration)     
        x_matrix = dotx_matrix + x_matrix
        
        #==============================================================================
        # Append the list : unemploy_rate
        unemploy_rate.append(1-L_m/L_mu)
        unemployed_tot.append(unemployed/(self.grid.width*self.grid.height))         
        W_a_ev.append(compute_Wa())
        W_m_ev.append(compute_Wm())
        #==============================================================================
        # Update wages
        for i in range(self.grid.height):                                       # Change all agents
            for j in range(self.grid.width):
                pos = (i, j)
                cellmates = self.grid.get_cell_list_contents(pos)
                if origin_list[i][j] == 2:
                    cellmates[0].wage = compute_Wm()                            # Manufacturing wages
                elif origin_list[i][j] == 1:                                    
                    cellmates[0].wage = 0                                       # No wages for unemployed
                else:
                    cellmates[0].wage = compute_Wa()                            # Agricultural wages
                    
    #==============================================================================
    # Run a step/an iteration for the model 
    # 
    # i = iteration number
    #==============================================================================    
    def step(self, i):
        # self.datacollector.collect(self)
        self.schedule.step()                # step all agent
        self.update_data()                  # update data model      
        
        global origin_list
        global migra_list
        global x_matrix
        global dotx_matrix
        global debug
        if debug:
            plot_list(origin_list, str(i))
            plot_list(x_matrix, "x"+str(i))
            plot_list(dotx_matrix, "dotx"+str(i))
        migra_list = np.zeros((self.grid.height, self.grid.width))          # Reset migration list
        dotx_matrix = np.zeros((self.grid.height, self.grid.width))  

#==============================================================================
# MESA AGENT
# 
# Agent = agent id (number from 0 to Pop-1)
#==============================================================================
class Harris_Todaro_Agent(Agent):

    #==============================================================================
    # MESA AGENT initialization
    # 
    #============================================================================== 
    def __init__(self, unique_id, model):
        super().__init__(unique_id, model)      # init agent
        x = int(unique_id/model.grid.width)
        y = unique_id%model.grid.width
        global origin_list                      # import origin list
        if origin_list[x, y] == 0:              # wage = rural agent
            self.wage = compute_Wa()
        else:                                   # wage = employed urban agent
            self.wage = compute_Wm()

    #==============================================================================
    # update agent wage
    # 
    #============================================================================== 
    def update_wage(self):
        global origin_list
        global migra_list
        global x_matrix
        global dotx_matrix
        global weights_matrix_matrix
        global width
        global height
        global x_ev

        Wm = compute_Wm()
        Wa = compute_Wa()

        social_influence = 0
        for i in range(height):
            for j in range(width):
                social_influence += weights_matrix_matrix[self.pos[0]][self.pos[1]][i][j] * (x_matrix[i,j]-x_matrix[self.pos[0]][self.pos[1]])    

        dotx_matrix[self.pos[0]][self.pos[1]] = a * x_matrix[self.pos[0]][self.pos[1]] + f * social_influence + b * ((L_m/L_mu) * Wm - Wa) + c * (self.wage - Wa)

        prob = np.abs(x_matrix[self.pos[0]][self.pos[1]]+dotx_matrix[self.pos[0]][self.pos[1]] )/(np.abs(x_matrix[self.pos[0]][self.pos[1]]+dotx_matrix[self.pos[0]][self.pos[1]] )+beta)
        
        if random.random() < prob:
            if x_matrix[self.pos[0]][self.pos[1]]+dotx_matrix[self.pos[0]][self.pos[1]] > 0:
                migra_list[self.pos[0]][self.pos[1]]=2
            else:
                migra_list[self.pos[0]][self.pos[1]]=0
        else:
            if origin_list[self.pos[0]][self.pos[1]]==2:
                migra_list[self.pos[0]][self.pos[1]]=2
            if origin_list[self.pos[0]][self.pos[1]]==1:
                migra_list[self.pos[0]][self.pos[1]]=2                              
      
    #==============================================================================
    # Run a step/an iteration for the agent 
    # 
    #==============================================================================    
    def step(self):
        # self.move()
        self.update_wage()                                  # update agent wage









#******************************************************************************
#******************************************************************************
# MAIN

if __name__ == '__main__': 

    test = 0
    f_param = [0, 0.0005, 0.001, 0.002, 0.004, 0.008, 0.016, 0.032]
    b_param = [0, 0.0025,0.005, 0.0075, 0.01,0.02,0.04,0.06]    
    a_param = [0, 0.00005, 0.0002, 0.0008, 0.00032,0.00128,0.00512,0.02048]
    c_param = [0, 0.005, 0.01, 0.02, 0.04, 0.08, 0.16, 0.32]
    beta_param = [0, 1, 2, 3, 4, 5, 6, 7]
    find_job_param = [0, 0.2, 0.4, 0.6, 0.8, 1]
    W_m_param = [0.6, 0.65, 0.7, 0.75, 0.8, 0.85, 0.9, 0.95, 1]
    L_mu_param = [0.1, 0.15, 0.2, 0.25, 0.3, 0.35, 0.4, 0.45, 0.5]
    A_a_param = [0.6, 0.7, 0.8, 0.9, 1, 1.1, 1.2, 1.3, 1.4]
    A_m_param = [0.6, 0.7, 0.8, 0.9, 1, 1.1, 1.2, 1.3, 1.4]
    alpha_param = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]
    phi_param = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]
    constant_param = [1, 1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8]
    for constant in constant_param:        
        #==============================================================================
        # Initialize all parameters and initial conditions
        # 
        #==============================================================================
        A_a = 1     # Parametric constant in rural sector production function       initial = 1
        A_m = 1     # Parametric constant in urban sector production function       initial = 1
        phi = 0.3   # Parametric constant in rural sector production function       initial = 0.3
        alpha = 0.7 # Parametric constant in urban sector production function       initial = 0.7
        sigma = 1   # Parametric constant in trade                                  initial = 1
        rho = 1     # Parametric constant in trade                                  initial = 1
        
        debug = False
        concensus = False
        find_job = 0.2   
        #constant = 1
        random_unemployment = True
        if concensus:
            a = 0.002
            f = 0.004
            b = 0.02
            c = 0.01
            beta = 2
            social_link = 0.1
            sparse_factor = 0.08
        else:
            a = 0.0008
            f = 0.001
            b = 0.02
            c = 0.01
            beta = 3
            social_link = 0.1
            sparse_factor = 0.09





        W_m = 0.8   # Minimum wage in urban sector                                  initial = 0.8
        L_mu = 0.2      # labor input in urban sector (employed and unemployed)     initial = 0.2                   
        L_a = 1 - L_mu  # labor input in rural sector                               initial = 1 - L_mu = 0.8
        L_m = 0.2       # labor input in manufacturing jobs                         initial = 0.2
        L_m = L_mu
        L_mu_share = [L_mu]           # urbanization (urban/total population share)                                 LIST = 1 value per iteration
        unemploy_rate = [1-L_m/L_mu]  # urban unemployment rate                     initial = [1-L_m/L_mu] = 0      LIST = 1 value per iteration
        unemployed_tot = [0]  # urban unemployment rate                     initial = [1-L_m/L_mu] = 0      LIST = 1 value per iteration
        Lm_opt = compute_Lm_opt()     # optimal manufacturing jobs

        width = 30      # size of the population = width x height = 300 x 300 = 90000
        height = 30
        iterations = 300     # number of iterations of the model


        #==============================================================================
        # Initialize origin and migration matrices
        # 
        #==============================================================================
        origin_list, migra_list, x_matrix, dotx_matrix, weights_matrix_matrix = init_list(height,width)       # Initialization of population
        plot_list(origin_list, 'origin')                        # plot initial population (origin list)


        #==============================================================================
        # Initialize model and agents
        # 
        #==============================================================================
        model = MigrationModel(height*width, width, height)     # Initialization of MESA MODEL
        W_m_ev = [compute_Wm()]
        W_a_ev = [compute_Wa()]


        #==============================================================================
        # Iterate the model
        # 
        #==============================================================================
        print("Running model : ")
        for i in tqdm(range(iterations)):                       # Running the MESA MODEL n-times
            model.step(i)   # running the MESA MODEL for 1 iteration
    # results of initial conditions

        exp_urban_w = [W_m_ev[0]*(1-unemploy_rate[0])]
        for i in range(len(W_m_ev)-1):
            exp_urban_w.append(W_m_ev[i+1]*(1-unemploy_rate[i+1]))
        #==============================================================================
        # Plots and prints
        # 
        #==============================================================================
        plot_grid(origin_list)
        plot_line(L_mu_share, "urban share")
        plot_line(unemploy_rate, "unemployment_rate")
        plt.figure(dpi = 500)
        plt.plot(W_m_ev, label="manufacturing wage")
        plt.plot(exp_urban_w, label="expected urban wage")
        plt.plot(W_a_ev, label="agricultural wage")
        plt.xlabel("simulation steps")
        plt.ylabel("wage")
        plt.title("wages")
        plt.legend()
        plt.grid()
        plt.savefig('wages.png')
        plt.show() 
        if debug:
            print("Video population : ")
            gen_video(iterations,"result")
            print("Creating images : ")
            for i in tqdm(range(iterations+1)):
                plot_line_dot(L_mu_share, "urban_share",i)
                plot_line_dot(unemploy_rate, "unemployment_rate",i)
            print("Video urban share : ")
            gen_video(iterations,"urban_share")
            print("Video unemployment rate : ")
            gen_video(iterations,"unemployment_rate")
            print("Video propensity")
            gen_video(iterations, "x")
            print("Video propensity derivative")
            gen_video(iterations, "dotx")
        # print(L_mu_share)
        # print(unemploy_rate)
        with open('constant.txt', 'a') as file:
            # Write the data to the file
            file.write("W_m_" + str(test) + " = " + str(W_m_ev) + "\n")
            file.write("exp_urban_w_" + str(test) + " = " + str(exp_urban_w) + "\n")
            file.write("W_a_" + str(test) + " = " + str(W_a_ev) + "\n")
            file.write("L_mu_share_" + str(test) + " = " + str(L_mu_share) + "\n")
            file.write("unemploy_rate_" + str(test) + " = " + str(unemploy_rate) + "\n")
        test +=1






